/*******************************************************************************
Copyright © 2023 Sebastian Roith

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*******************************************************************************/

(function(window) {

function isOnlyCommands({
    tokens      // Quellcode als Array von Token-Instanzen
}) {
    const realTokens = tokens.filter(token => (!(token instanceof TokenBracketOpen) && !(token instanceof TokenBracketClose) && !(token instanceof TokenBlockComment) && !(token instanceof TokenLineComment) && !(token instanceof TokenWhitespace )&& !(token instanceof TokenIndent)));
    let wrongTokens = realTokens.filter(token => !(token instanceof TokenCommand));
    if (wrongTokens.length)
        return { accept: false, msg: 'Bei dieser Aufgabe sollen nur Algorithmen geschrieben werden, die aus Sequenzen und Anweisungen bestehen. Also keine selbst definierten Anweisungen und auch sonst nichts, was du eigentlich noch nicht gelernt hast.', line: wrongTokens[0].line };
    return { accept: true };
}

function isOnlyCommandsAndUserdefined(tokens, allowVertieft) {
    const realTokens = tokens.filter(token => (!(token instanceof TokenBracketOpen) && !(token instanceof TokenBracketClose) && !(token instanceof TokenBlockComment) && !(token instanceof TokenLineComment) && !(token instanceof TokenWhitespace )&& !(token instanceof TokenIndent)));
    const allowed = [TokenCommand, TokenDefinition];
    if (allowVertieft)
        allowed.push(TokenRunFast);
    const wrongTokens = realTokens.filter(token => !allowed.reduce((prev, curr) => (prev || (token instanceof curr)), false));
    return wrongTokens.length;
}

/**
* Namen aller vom Nutzer definierten Funktionen herausfinden
* @return Array     Funktionsnamen in Reihenfolge der Definitionen
*/
function getUserDefinedNames(compiled) {
    const defs = compiled.getContentByClass(SourceCodeDefinition);
    return defs.map(def => def.name);
}

/**
* Die SourceCodeDefinition-Instanz finden, die die selbstdefinierte Anweisung namens fktName festlegt
* @return SourceCodeDefinition      Gefundene Instanz (oder null)
*/
function getUserDefinedByName(
    compiled,       // SourceCodePart-Instanz, die kompletten Quellcode repräsentiert
    fktName         // Name der fraglichen selbstdefinierten Anweisung
) {
    const def = compiled.getContentByClass(SourceCodeDefinition).filter(def => (def.name == fktName));
    return def.length ? def[0] : null;
}

/**
* Überprüfen, ob eine selbstdefinierte Anweisung exakt den vorgegebenen Inhalt hat
* @return Object    Abweichung (angelehnt an Rückgabewert von SourceCodePart.prototype.compare) oder null
*/
function cmpUserDefined(
    compiled,       // SourceCodePart-Instanz, die kompletten Quellcode repräsentiert
    fktName,        // Name der zu überprüfenden selbstdefinierten Anweisung
    code,           // Genaue Vorgabe, wie die Anweisung programmiert sein muss
    forceRunFast    // Optional: Führt Verwendung des Schlüsselwortes VERTIEFT zu einem Fehler (false), fehlende Verwendung zu einem Fehler (true) oder wird das nicht geprüft (Standardwert undefined)?
) {
    const def = getUserDefinedByName(compiled, fktName);
    if (!def)
        return { index: -1, actualValue: undefined, plannedValue: fktName, type: 'notDefined' };
    if ((forceRunFast === true) && def.isSlow)
        return { index: -1, type: 'runFast', actualValue: false, plannedValue: true };
    if ((forceRunFast === false) && !def.isSlow)
        return { index: -1, type: 'runFast', actualValue: true, plannedValue: false };
    const soll = SourceCodePart.tokensToSequenz(Token.tokenize(code), true);
    return def.body.compare(soll);
}

/**
* Neue selbst definierter Funktion zurückweisen, wenn sie nicht den Anforderungen entspricht
* @return Object    null (für keine Beanstandung) oder passende Rückgabewert für checkCode bzw. checkTermination
*/
function ensureUserDefinedAdd(
    compiled,       // SourceCodePart-Instanz, die kompletten Quellcode repräsentiert
    fktName,        // Name der zu überprüfenden selbstdefinierten Anweisung
    code,           // Genaue Vorgabe, wie die Anweisung programmiert sein muss
    forceRunFast    // Optional: Führt Verwendung des Schlüsselwortes VERTIEFT zu einem Fehler (false), fehlende Verwendung zu einem Fehler (true) oder wird das nicht geprüft (Standardwert undefined)?
) {
    const cmp = cmpUserDefined(compiled, fktName, code, forceRunFast);
    if (!cmp)
        return null;
    const result = { accept: false, fine: 1 };
    switch (cmp.type) {
        case 'notDefined':
            const noDef = compiled.content.filter(obj => !(obj instanceof SourceCodeDefinition));
            if (noDef.length)
                result.line = noDef[0].line;
            result.msg = 'Die selbstdefinierte Anweisung ' + fktName + ' fehlt noch immer…';
            return result;
        case 'runFast':
            const def = getUserDefinedByName(compiled, fktName);
            if (def instanceof SourceCodePart)
                result.line = def.line;
            result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' ';
            result.msg += forceRunFast ? 'fehlt das Schlüsselwort VERTIEFT' : 'muss das Schlüsselwort VERTIEFT weg';
            return result;
        case 'item':
            if (result.actualValue instanceof SourceCodePart)
                result.line = result.actualValue.line;
            result.msg = 'Gut, nun gibt es eine selbstdefinierte Anweisung ' + fktName + '. Aber was sie tut, passt noch nicht…';
            return result;
        default:
            result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' stimmt etwas nicht…';
            if (result.actualValue instanceof SourceCodePart)
                result.line = result.actualValue.line;
            return result;
    }
}

/**
* Änderungen an selbst definierter Funktion zurückweisen
* @return Object    null (für keine Beanstandung) oder passende Rückgabewert für checkCode bzw. checkTermination
*/
function ensureUserDefinedUnchanged(
    compiled,       // SourceCodePart-Instanz, die kompletten Quellcode repräsentiert
    fktName,        // Name der zu überprüfenden selbstdefinierten Anweisung
    code,           // Genaue Vorgabe, wie die Anweisung programmiert sein muss
    forceRunFast    // Optional: Führt Verwendung des Schlüsselwortes VERTIEFT zu einem Fehler (false), fehlende Verwendung zu einem Fehler (true) oder wird das nicht geprüft (Standardwert undefined)?
) {
    const cmp = cmpUserDefined(compiled, fktName, code, forceRunFast);
    if (!cmp)
        return null;
    const result = { accept: false, fine: 1 };
    switch (cmp.type) {
        case 'notDefined':
            const noDef = compiled.content.filter(obj => !(obj instanceof SourceCodeDefinition));
            if (noDef.length)
                result.line = noDef[0].line;
            result.msg = 'Die selbstdefinierte Anweisung ' + fktName + ' soll es weiterhin geben…';
            return result;
        case 'runFast':
            const def = getUserDefinedByName(compiled, fktName);
            if (def instanceof SourceCodePart)
                result.line = def.line;
            result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' ';
            result.msg += forceRunFast ? 'fehlt das Schlüsselwort VERTIEFT' : 'muss das Schlüsselwort VERTIEFT weg';
            return result;
        case 'item':
            if (result.actualValue instanceof SourceCodePart)
                result.line = result.actualValue.line;
            result.msg = 'Eine Veränderung der selbstdefinierten Anweisung ' + fktName + ' ist hier falsch';
            return result;
        default:
            result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' stimmt etwas nicht…';
            if (result.actualValue instanceof SourceCodePart)
                result.line = result.actualValue.line;
            return result;
    }
}

/**
* Überprüfen, ob selbst definierte Funktion auf gewünschte Weise abgeändert wurde
* @return Object    null (für keine Beanstandung) oder passende Rückgabewert für checkCode bzw. checkTermination
*/
function ensureUserDefinedChanged(
    compiled,       // SourceCodePart-Instanz, die kompletten Quellcode repräsentiert
    fktName,        // Name der zu überprüfenden selbstdefinierten Anweisung
    codeUnchanged,  // Genaue Vorgabe, wie die Anweisung vor der Änderung programmiert ist
    codeChanged,    // Genaue Vorgabe, wie die abgeänderte Anweisung programmiert sein muss
    forceRunFast    // Optional: Führt Verwendung des Schlüsselwortes VERTIEFT zu einem Fehler (false), fehlende Verwendung zu einem Fehler (true) oder wird das nicht geprüft (Standardwert undefined)?
) {
    const cmp = cmpUserDefined(compiled, fktName, codeChanged, forceRunFast);
    if (!cmp)
        return null;
    const result = { accept: false, fine: 1 };
    switch (cmp.type) {
        case 'notDefined':
            const noDef = compiled.content.filter(obj => !(obj instanceof SourceCodeDefinition));
            if (noDef.length)
                result.line = noDef[0].line;
            result.msg = 'Die selbstdefinierte Anweisung ' + fktName + ' soll es weiterhin geben…';
            return result;
        case 'runFast':
            const def = getUserDefinedByName(compiled, fktName);
            if (def instanceof SourceCodePart)
                result.line = def.line;
            result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' ';
            result.msg += forceRunFast ? 'fehlt das Schlüsselwort VERTIEFT' : 'muss das Schlüsselwort VERTIEFT weg';
            return result;
        case 'item':
            const unchanged = cmpUserDefined(compiled, fktName, codeUnchanged) ? false : true;
            if (unchanged) {
                const def = getUserDefinedByName(compiled, fktName);
                if (def instanceof SourceCodePart)
                    result.line = def.line;
                result.msg = 'Um die Aufgabe zu lösen, musst du die selbstdefinierte Anweisung ' + fktName + ' verändern';
            return result;
            }
            if (result.actualValue instanceof SourceCodePart)
                result.line = result.actualValue.line;
            else {
                const def = getUserDefinedByName(compiled, fktName);
                if (def instanceof SourceCodePart)
                    result.line = def.line;
            }
            result.msg = 'Richtig, die selbstdefinierte Anweisung ' + fktName + ' muss verändert werden. Aber nicht so…';
            return result;
        default:
            result.msg = 'Bei der selbstdefinierten Anweisung ' + fktName + ' stimmt etwas nicht…';
            if (result.actualValue instanceof SourceCodePart)
                result.line = result.actualValue.line;
            return result;
    }
}

/**
* Betrachtet nur die TokenCommand-Instanzen und überprüft, ob es die geforderten in der richtigen Reihenfolge sind
* @return Object    null (für keine Beanstandung) oder passende Rückgabewert für checkCode bzw. checkTermination
*/
function compareOnlyCommands(
    tokens,     // Alle Token-Instanzen, die aus dem Quellcode resultieren
    planed,     // Array aus Anweisungsnamen, die in dieser Reihenfolge im Quellcode stehen sollen
    strHint     // Optional: Zusatztext für Fehlermeldung
) {
    const result = { accept: false, fine: 1, msg: strHint ? (strHint + ': ') : '' };
    const commands = tokens.filter(token => (token instanceof TokenCommand));
    for (let i = 0; i < commands.length; ++i) {
        if (commands[i].name == planed[i])
            continue;
        result.line = commands[i].line;
        if (i < planed.length)
            result.msg += 'Statt ' + commands[i].name + ' muss hier ' +  planed[i] + ' stehen';
        else
            result.msg += 'Die zusätzliche Anweisung ' + commands[i].name + ' muss weg';
        return result;
    }
    if (commands.length < planed.length) {
        result.msg += 'Am Ende fehlt die Anweisung ' + planed[commands.length];
        if (commands.length + 1 < planed.length)
            result.msg += ' uvm.';
        result.line = tokens.length ? tokens[tokens.length - 1].line : 0;
        return result;
    }
    return null;
}

/**
* Sicherstellen, dass bestimmte Anzahl an Anweisungsaufrufen im Quelltext stehen
* @return Object    null (für keine Beanstandung) oder passende Rückgabewert für checkCode bzw. checkTermination
*/
function ensureNumberOfFunctionCalls(
    tokens,     // Alle Token-Instanzen, die aus dem Quellcode resultieren
    fktName,    // Name der fraglichen Anweisung
    offset,     // Korrektur der gezählten Funktionsaufrufe (beispielsweise wegen deren Definition)
    min,        // Geforderte Mindestzahl an Funktionsaufrufen
    max         // Geforderte Maximalanzahl an Funktionsaufrufen (Infinity ist erlaubt)
) {
    const commands = tokens.filter(token => ((token instanceof TokenCommand) && (token.name == fktName)));
    const numberOfCalls = commands.length - offset;
    if (numberOfCalls > max)
        return {
            accept: false,
            fine: 1,
            line: commands[max + offset].line,
            msg: 'So viele Aufrufe von ' + fktName + ' sind nicht erforderlich'
        };
    if (numberOfCalls < min)
        return {
            accept: false,
            fine: 1,
            line: tokens.length ? tokens[tokens.length - 1].line : 1,
            msg: 'Es müsste' + (numberOfCalls ? ' mehr' : '') + ' Aufrufe von ' + fktName + ' geben…'
        };
    return null;
}

function checkWorld(
    worldStat,
    arrCounts       // Flaches Array mit Vorgaben, wie viele Plättchen auf dem jeweiligen Feld liegen sollen (Standardwert 0)
) {
    if (!(arrCounts instanceof Array))
        arrCounts = [];
    let wrong = 0;
    let index = 0;
    for (let x = 0; x < worldStat.width; ++x) {
        for (let y = 0; y < worldStat.height; ++y) {
            if (worldStat.read(x, y) != ((index < arrCounts.length) ? arrCounts[index] : 0))
                ++wrong;
            ++index;
        }
    }
    return wrong;
}

function worldChanged(worldStat) {
    const oldWorld = new RobotSynchronizedWorld();
    oldWorld.init(this.world);
    return !oldWorld.sameWorld(worldStat);
}

/**
* Erstellt eine für checkTermination geeignete Funktion, die überprüft, ob die Ergebniswelt den Vorgaben entspricht, und einen entsprechenden Rückgabewert erstellt
* @return Function      Für checkTermination geeignete Funktion
*/
function createWorldChecker(
    arrCounts,      // Flaches Array mit Vorgaben, wie viele Plättchen auf dem jeweiligen Feld liegen sollen (Standardwert 0)
    isNoChange      // Optional: Beschreibt arrCounts auch die Ergebniswelt vor einer Programmänderung?
) {
    return function({ worldStat, result }) {
        if (result != RobotSimulator.RESULT_DONE)
            return { accept: false, fine: 1 };
        const wrong = checkWorld(worldStat, arrCounts);
        const nameWrong = ['', 'einem Feld', 'zwei Feldern', 'drei Feldern', 'vier Feldern', 'fünf Feldern', 'sechs Feldern', 'sieben Feldern', 'acht Feldern', 'neun Feldern', 'zehn Feldern', 'elf Feldern', 'zwölf Feldern'];
        if (wrong) {
            const strWrong = ((wrong < nameWrong.length) ? nameWrong[wrong] : (wrong + ' Feldern'));
            const result = { accept: false, fine: 1, msg: 'Bei ' + strWrong + ' passt die Plättchenanzahl nicht' };
            if (isNoChange)
                result.msg = 'Du hast zu viel verändert: ' + result.msg + ' mehr.';
            else
                result.msg += '.';
            return result;
        }
        return { unlock: true, accept: true };
    }
}

const LEVELS = [
    {
        // Angezeigter Arbeitsauftrag (HTML erlaubt)
        task: 'Eine vordefinierte Anweisung ist bekanntlich <code>linksrum</code>. Bringe nun dem Roboter die Anweisung <code>rechtsrum</code> bei. Insgesamt soll sie bewirken, dass der Roboter um 90° nach rechts gedreht dasteht. Dazu muss er sich entsprechend weit nach links drehen. Verzichte zunächst auf einen Aufruf dieser Anweisung oder sonst was:',

        // Welt und Roboterstartposition als Maschinencode
        world: '031100000',

        // Quelltext, der anfangs im Editor steht
        code: '# Roboter dreht sich insgesamt um 90° nach rechts\n',

        // Ist das Bearbeiten des Quelltextes gesperrt? (Optional, Standardwert false)
        readonly: false,

        // Muss - ggf. zusätzlich zur korrekten Fragenbeantwortung - checkCode oder checkTermination unlock=true liefern, bevor Level abgeschlossen werden kann? (Optional, Standardwert false)
        locked: true,

        // Optional: Überprüft, ob ein syntaktisch korrekter Quellcode ausgeführt werden darf; Rückgabewert true für ja, sonst { accept: "Ausführung erlauben?", line: "Zeilennummer eines Fehlers (optional)", msg: "Fehlermeldung (optional)", fine: "Als einen Fehlversuch werten? (optional)", unlock: "Level als abgeschlossen behandeln? (optional)" }
        checkCode({
            code,       // In Editor eingegebener Code
            tokens,     // Quellcode als Array von Token-Instanzen
            compiled,   // Quelltextbaum als SourceCodePart-Instanz
            worldDef    // Maschinencode der Welt zu Beginn
        }) {
            if (isOnlyCommandsAndUserdefined(tokens))
                return { accept: false, fine: 1, msg: 'Du sollst nur die Anweisung rechtsrum schreiben, ohne erst später erklärte Kontrollstrukturen zu nutzen' };
            const defs = getUserDefinedNames(compiled);
            if (defs.length > 1)
                return { accept: false, fine: 1, msg: 'Es soll keine zusätzliche Anweisung geben, nur die Anweisung rechtsrum' };
            if (defs.indexOf('rechtsrum') == -1)
                return { accept: false, fine: 1, msg: 'Es gibt noch immer keine Anweisung namens rechtsrum' };

            const cmpRechtsrum = ensureUserDefinedAdd(compiled, 'rechtsrum', 'linksrum\nlinksrum\nlinksrum');
            if (cmpRechtsrum)
                return cmpRechtsrum;

            const cmpCalls = ensureNumberOfFunctionCalls(tokens, 'rechtsrum', 1, 0, 0);
            if (cmpCalls)
                return cmpCalls;

            const cmpCommands = compareOnlyCommands(tokens, ['rechtsrum', 'linksrum', 'linksrum', 'linksrum']);
            if (cmpCommands)
                return cmpCommands;

            return { accept: true };
        },

        // Optional: Reagiert auf ein Ende der Algorithmusausführung; Rückgabewert wie checkCode
        checkTermination({
            code,       // In Editor eingegebener Code
            line,       // Zuletzt ausgeführte Zeilennummer
            result,     // Grund für das Ende: undefined bei manuellem Abbruch, sonst RobotSimulator.RESULT_*
            worldDef,   // Maschinencode der Welt zu Beginn
            worldStat,  // Present robot's world after execution (RobotWorld instance)
            robotX,     // Location of robot after execution in x direction (north)
            robotY,     // Location of robot after execution in y direction (east)
            robotDirection      // Viewing direction of robot after execution
        }) {
            return createWorldChecker([0], true)({ code, line, result, worldDef, worldStat, robotX, robotY, robotDirection });
        },

        // Wird falsches Beantworten der Frage als Fehlversuch gewertet? (Optional, Standardwert false)
        questionWrongFine: 0,

        // Falls erst eine Frage richtig beantwortet werden muss: Welche?
        question: null
    },
    {
        task: 'Deine selbstdefinierte Anweisung <code>rechtsrum</code> passt. Teste nun, ob sie funktioniert.',
        world: '031100000',
        code: '# Roboter dreht sich insgesamt um 90° nach rechts\nLERNE rechtsrum\n    linksrum\n    linksrum\n    linksrum',
        keepCode: true,
        locked: true,
        checkCode({
            code,       // In Editor eingegebener Code
            tokens,     // Quellcode als Array von Token-Instanzen
            compiled,   // Quelltextbaum als SourceCodePart-Instanz
            worldDef    // Maschinencode der Welt zu Beginn
        }) {
            if (isOnlyCommandsAndUserdefined(tokens))
                return { accept: false, fine: 1, msg: 'Außer Anweisungen, Sequenzen und der selbstdefinierten Anweisung soll es nichts weiteres geben' };
            const defs = getUserDefinedNames(compiled);
            if (defs.length > 1)
                return { accept: false, fine: 1, msg: 'Es soll keine zusätzliche Anweisung geben, nur die Anweisung rechtsrum' };
            if (defs.indexOf('rechtsrum') == -1)
                return { accept: false, fine: 1, msg: 'Es gibt noch immer keine Anweisung namens rechtsrum' };

            const cmpErwartet = ensureUserDefinedChanged(compiled, 'rechtsrum', 'linksrum\nlinksrum\nlinksrum', 'linksrum\nlinksrum\nlinksrum\nrechtsrum');
            if (!cmpErwartet)
                return { accept: false, fine: 1, line: tokens[tokens.length - 1].line, msg: 'Der Aufruf von rechtsrum darf nicht eingerückt sein' }

            const cmpRechtsrum = ensureUserDefinedUnchanged(compiled, 'rechtsrum', 'linksrum\nlinksrum\nlinksrum');
            if (cmpRechtsrum)
                return cmpRechtsrum;

            const cmpCalls = ensureNumberOfFunctionCalls(tokens, 'rechtsrum', 1, 1, 1);
            if (cmpCalls)
                return { accept: false, fine: 1, line: cmpCalls.line, msg: 'Um sie testen zu können, musst du die Anweisung rechtsrum genau einmal aufrufen' };

            const cmpCommands = compareOnlyCommands(tokens, ['rechtsrum', 'linksrum', 'linksrum', 'linksrum', 'rechtsrum']);
            if (cmpCommands)
                return cmpCommands;

            return { accept: true };
        },
        checkTermination: createWorldChecker([0], true)
    },
    {
        task: 'Neue Aufgabe: Erfinde eine selbstdefinierte Anweisung <code>plus3</code>, die den Roboter drei Plättchen hinlegen lässt. Teste gleich, ob sie funktioniert.',
        world: '031100000',
        code: '# Roboter legt drei Plättchen auf sein aktuelles Feld\n',
        locked: true,
        checkCode({
            code,       // In Editor eingegebener Code
            tokens,     // Quellcode als Array von Token-Instanzen
            compiled,   // Quelltextbaum als SourceCodePart-Instanz
            worldDef    // Maschinencode der Welt zu Beginn
        }) {
            if (isOnlyCommandsAndUserdefined(tokens))
                return { accept: false, fine: 1, msg: 'Außer Anweisungen, Sequenzen und der selbstdefinierten Anweisung soll es nichts weiteres geben' };
            const defs = getUserDefinedNames(compiled);
            if (defs.length > 1)
                return { accept: false, fine: 1, msg: 'Es soll keine zusätzliche Anweisung geben, nur die Anweisung plus3' };
            if (defs.indexOf('plus3') == -1)
                return { accept: false, fine: 1, msg: 'Es gibt noch immer keine Anweisung namens plus3' };

            const cmpPlus3 = ensureUserDefinedAdd(compiled, 'plus3', 'plus\nplus\nplus');
            if (cmpPlus3)
                return cmpPlus3;

            const cmpCalls = ensureNumberOfFunctionCalls(tokens, 'plus3', 1, 1, 1);
            if (cmpCalls)
                return { accept: false, fine: 1, line: cmpCalls.line, msg: 'Um sie testen zu können, musst du die Anweisung plus3 genau einmal aufrufen' };

            return { accept: true };
        },
        checkTermination: createWorldChecker([3])
    },
    {
        task: 'Jetzt noch eine selbstdefinierte Anweisung <code>plus12</code>, die den Roboter zwölf Plättchen hinlegen lässt. Nutze die selbstdefinierte Anweisung <code>plus3</code>, damit der Algorithmus kurz bleibt. Teste gleich, ob <code>plus12</code> funktioniert.',
        world: '031100000',
        code: '# Roboter legt drei Plättchen auf sein aktuelles Feld\nLERNE plus3\n    plus\n    plus\n    plus\n\nplus3',
        keepCode: true,
        locked: true,
        checkCode({
            code,       // In Editor eingegebener Code
            tokens,     // Quellcode als Array von Token-Instanzen
            compiled,   // Quelltextbaum als SourceCodePart-Instanz
            worldDef    // Maschinencode der Welt zu Beginn
        }) {
            if (isOnlyCommandsAndUserdefined(tokens))
                return { accept: false, fine: 1, msg: 'Außer Anweisungen, Sequenzen und der selbstdefinierten Anweisung soll es nichts weiteres geben' };
            const defs = getUserDefinedNames(compiled);
            if (defs.length > 2)
                return { accept: false, fine: 1, msg: 'Außer den Anweisungen plus3 und plus12 soll es keine weiteren selbstdefinierten Anweisungen geben' };
            if (defs.indexOf('plus12') == -1)
                return { accept: false, fine: 1, msg: 'Es gibt noch immer keine Anweisung namens plus12' };

            const cmpPlus3 = ensureUserDefinedUnchanged(compiled, 'plus3', 'plus\nplus\nplus');
            if (cmpPlus3)
                return cmpPlus3;

            const cmpPlus12 = ensureUserDefinedAdd(compiled, 'plus12', 'plus3\nplus3\nplus3\nplus3');
            if (cmpPlus12)
                return cmpPlus12;

            const cmpCalls = (ensureNumberOfFunctionCalls(tokens, 'plus3', 5, 0, 0) || ensureNumberOfFunctionCalls(tokens, 'plus12', 1, 1, 1));
            if (cmpCalls)
                return { accept: false, fine: 1, line: cmpCalls.line, msg: 'Um sie testen zu können, musst du nur die Anweisung plus12 genau einmal aufrufen' };

            return { accept: true };
        },
        checkTermination: createWorldChecker([12])
    },
    {
        task: 'Ergänze nun noch eine selbstdefinierte Anweisung <code>plus15</code>, die den Roboter 15 Plättchen hinlegen lässt. Nutze geschickt die anderen selbstdefinierten Anweisungen, um die Definition von <code>plus15</code> so kurz wie möglich zu halten. Und teste nur <code>plus15</code>.',
        world: '031100000',
        code: '# Roboter legt drei Plättchen auf sein aktuelles Feld\nLERNE plus3\n    plus\n    plus\n    plus\n\n# Roboter legt zwölf Plättchen auf sein aktuelles Feld\nLERNE plus12\n    plus3\n    plus3\n    plus3\n    plus3\n\nplus12',
        keepCode: true,
        locked: true,
        checkCode({
            code,       // In Editor eingegebener Code
            tokens,     // Quellcode als Array von Token-Instanzen
            compiled,   // Quelltextbaum als SourceCodePart-Instanz
            worldDef    // Maschinencode der Welt zu Beginn
        }) {
            if (isOnlyCommandsAndUserdefined(tokens))
                return { accept: false, fine: 1, msg: 'Außer Anweisungen, Sequenzen und der selbstdefinierten Anweisung soll es nichts weiteres geben' };
            const defs = getUserDefinedNames(compiled);
            if (defs.length > 3)
                return { accept: false, fine: 1, msg: 'Außer den Anweisungen plus3, plus12 und plus15 soll es keine weiteren selbstdefinierten Anweisungen geben' };

            const cmpPlusX = (ensureUserDefinedUnchanged(compiled, 'plus3', 'plus\nplus\nplus') || ensureUserDefinedUnchanged(compiled, 'plus12', 'plus3\nplus3\nplus3\nplus3'));
            if (cmpPlusX)
                return cmpPlusX;

            const cmpPlus15a = ensureUserDefinedUnchanged(compiled, 'plus15', 'plus3\nplus12');
            const cmpPlus15b = ensureUserDefinedUnchanged(compiled, 'plus15', 'plus12\nplus3');
            if (cmpPlus15a && cmpPlus15b)
                return cmpPlus15b;

            const cmpCalls = (ensureNumberOfFunctionCalls(tokens, 'plus3', 6, 0, 0) || ensureNumberOfFunctionCalls(tokens, 'plus12', 2, 0, 0) || ensureNumberOfFunctionCalls(tokens, 'plus15', 1, 1, 1));
            if (cmpCalls)
                return { accept: false, fine: 1, line: cmpCalls.line, msg: 'Um sie testen zu können, musst du nur die Anweisung plus15 genau einmal aufrufen' };

            return { accept: true };
        },
        checkTermination: createWorldChecker([15])
    }
];

window.roboterCodingTask = {
    levels: LEVELS
};
})(window);
